"""
.. module:: ShutTUM
   :platform: Unix, Windows
   :synopsis: A utility API to easily interact with the ShutTUM dataset

.. moduleauthor:: Thore Goll <thore.goll@tum.de>

"""
import math
import numpy   as np


class Interpolation:
    """Interpolation lets you calculate intermediate values between two boarders, based on a given formula.

    .. image:: images/interpolation.svg       

    **ShutTUM**'s data is recorded at different frequencies. Images are timestamped at around 20 Hz, Imu measurements  
    at 160 Hz (exactly multiples per image), while Ground Truth data is clocked at 120 Hz. To 
    get relateable measurements, you must sometimes interpolate between two values.

    ShutTUM supports currently three interpolation methods, namely :any:`linear`, :any:`cubic` 
    and :any:`slerp`. However it is possible to define your own function and hand it over to interpolating 
    methods, such as  :any:`GroundTruth.interpolate` or :any:`Imu.interpolate`::

        def some_crazy_interpolation(a,b,t, a0, b0):
            return a + b - 17*t**2

        # Then use it like so
        t = 1.5  # [s] Time stamp at which to interpolate
        x = sequence.imu[0].interpolate(sequence, t, accelaration_interpolation=some_crazy_interpolation)

        # Now x is interpolated between the imu values closest to 1.5s with the crazy function
        
    """

    @staticmethod
    def linear(a, b, t, *_):
        r"""
        :param a: (float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the lower bound from which to interpolate
        :param b: (float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the upper bound to which to interpolate 
        :param t: (float) the value between 0 .. 1 of the interpolation
        :param _: further args are ignored but required in the function definition (see :any:`cubic`)
        :rtype: float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_

        Interpolate linearly between two points

        .. image:: images/interpolation-linear.svg

        .. math:: x(t) = \left (1 - t \right ) \cdot a + t \cdot b

        Example::

            # Find the average between two 3D vectors
            v1 = np.array([0,0,1])
            v2 = np.array([5,0,0])
            v  = Interpolation.linear(v1, v2, .5)   # np.array([2.5,0,0.5])

        """
        return (1 - t) * a + t * b

    @staticmethod
    def cubic(a, b, t, a0=None, b0=None):
        r"""
        :param a: (float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the lower bound from which to interpolate
        :param b: (float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the upper bound to which to interpolate
        :param t: (float) t: the value between 0 .. 1 of the interpolation
        :param a0: (float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the value before a, from which the tangent in a is calculated. If None, a0 = a
        :param b0: float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the value after b, from which the tangent in b is calculated. If None, b0 = b
        :rtype: float/`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_

        Interpolate cubically between two points. Note that a, b, a0 and a0 must all have the same shape.

        .. image:: images/interpolation-cubic.svg

        .. math:: 
            x(t) = a + \frac{1}{2} t \cdot \left ( b - a_0 + t \cdot \left (2 a_0 - 5 a + 4 b - b_0 + t \cdot \left (3 \cdot \left (a - b \right ) + b_0 - a_0 \right ) \right ) \right ) 
        """
        if a0 is None: a0 = a
        if b0 is None: b0 = b
        return a + 0.5 * t * (b - a0 + t * (2.0 * a0 - 5.0 * a + 4.0 * b - b0 + t * (3.0 * (a - b) + b0 - a0)))

    @staticmethod
    def slerp(quat0, quat1, t, *_):
        r"""
        :param quat0: (`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the lower rotational bound from which to interpolate (4x1 vector). Will be normalized
        :param quat1: (`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) the upper rotational bound to which to interpolate (4x1 vector). Will be normalized
        :param t: (`float`) the value between 0 .. 1 of the interpolation
        :param _: further args are ignored but required in the function definition (see :any:`cubic`)
        :rtype: `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_

        Interpolate spherically between two quaternions. This method is mostly used to interpolate rotations.
        Directly copied from `Transformation.py <http://www.lfd.uci.edu/~gohlke/code/transformations.py.html>`_
        """
        q0 = Interpolation._unit_vector(quat0[:4])
        q1 = Interpolation._unit_vector(quat1[:4])
        _EPS = np.finfo(float).eps * 4.0
        if t == 0.0:
            return q0
        elif t == 1.0:
            return q1
        d = np.dot(q0, q1)
        if abs(abs(d) - 1.0) < _EPS:
            return q0
        if d < 0.0:
            # invert rotation
            d = -d
            np.negative(q1, q1)
        angle = math.acos(d)
        if abs(angle) < _EPS:
            return q0
        isin = 1.0 / math.sin(angle)
        q0 *= math.sin((1.0 - t) * angle) * isin
        q1 *= math.sin(t * angle) * isin
        q0 += q1
        return q0

    @staticmethod
    def _unit_vector(data, axis=None, out=None):
        r"""
        Directly copied from `Transformation.py <http://www.lfd.uci.edu/~gohlke/code/transformations.py.html>`_
        """
        if out is None:
            data = np.array(data, dtype=np.float64, copy=True)
            if data.ndim == 1:
                data /= math.sqrt(np.dot(data, data))
                return data
        else:
            if out is not data:
                out[:] = np.array(data, copy=False)
            data = out
        length = np.atleast_1d(np.sum(data * data, axis))
        np.sqrt(length, length)
        if axis is not None:
            length = np.expand_dims(length, axis)
        data /= length
        if out is None:
            return data




import cv2
import numpy as np
import os.path as p
import transforms3d as tf
import ShutTUM
from collections import namedtuple


class Value(object):
    r"""
    A Value represents every measurement of a sensor taken in a specific time instance.
    A camera might capture an :any:`Image`, an IMU will record :any:`Imu` or the motion capture system 
    :any:`GroundTruth`.
     
    .. image:: images/frames.png 
     
    All these values have some common properties in the ShutTUM:
    
    1) They are single, time-discrete values (see :any:`stamp <ShutTUM.Value.stamp>`)
    2) They are related to a certain reference frame (see :any:`reference <ShutTUM.Value.reference>`)
    3) They all have a transformation from its reference frame towards that of ``"cam1"``
    
    
    Since every value has these three properties, you can achieve easily get the transformations between different  values. 
    Therefore the leftshift ``<<`` and rightshift ``>>`` operator has been overloaded. Both accept as their right parameter either:
    
    * a string indicating the desired reference to or from which to transform (e.g. ``"cam1"``, ``"cam2"``, ``"world"``, ``"imu"`` ...)
    * or another value, whose :any:`reference <ShutTUM.Value.reference>` property is used to determine the transform
    
    The direction of the "shift" means "How is the transformation from reference x to y?"::
    
        # assume we have an image and a ground truth value
        image = next(iter(sequence.cameras('global')))   # first frame
        gt    = next(sequence.mocap)                     # first ground truth
        
        # Since both Image and GroundTruth derive from Value, they have a reference ...
        print("Image %s is associated with reference %s" % (image, image.reference))
        print("Ground Truth %s is associated with reference %s" % (gt, gt.reference)) 
         
        # Since we know both transforms to "cam1", we can compute the transformation between the two
        P_img_wrld = image << gt     # pose of the image relative to the world 
        P_img_wrld = gt >> image     # same as line above
        P_img_wrld = gt >> "cam1"    # you can also specify the reference (on the right) as string directly
        # P_img_wrld = "cam1" << gt  # This fails, since the shift operators cannot be overloaded for strings in the first place
    
    """
    def __init__(self, sequence, stamp, reference):
        self._sequence = sequence
        self._stamp = stamp
        if reference not in sequence._refs:
            raise ValueError("[%s] Cannot find the reference %s" % (sequence, reference))
        self._reference = reference

    @property
    def _previous(self):
        return None

    def __eq__(self, other):
        if not isinstance(other, Value): return False
        return self._stamp == other._stamp and self.reference == other.reference

    @property
    def stamp(self):
        r"""
        The timestamp in seconds with decimal places. This stamp describes the exact moment this
        values has been recorded.
        """
        return self._stamp

    @property
    def reference(self):
        r"""
        The name as ``str`` of the reference frame, this value is associated with. 
        """
        return self._reference

    def dt(self, ifunknown=.001):
        r"""
        The time delta elapsed since the last :any:`Value` of this type.
        Note that this value varies depending on the concrete type of value. That means dt will be about 1/25 s
        for an :any:`Image` while around 1/120 s for a :any:`GroundTruth`. If a previous value is not present
        this will return time specified by ``ifunknown``.
        """
        previous = self._previous       # cache for performance
        if previous is None: return ifunknown
        else: return self.stamp - previous.stamp

    @property
    def _transform(self):
        r"""
        The transformation from ``"cam1"`` to this value`s :any:`reference <ShutTUM.Value.reference>` as 4x4 
        `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_ homogeneous matrix 
        """
        return np.array(self._sequence._refs[self._reference]['transform'])

    def __str__(self):
        return "%s (%s/%.2f)" % (type(self).__name__, self.reference, self.stamp)

    def __lshift__(self, parent):
        if isinstance(parent, str):
            if parent not in self._sequence._refs:
                raise ValueError("[%s] Cannot find the parent reference %s" % (self._sequence, parent))

            tparent = np.array(self._sequence._refs[parent]['transform'])

        elif isinstance(parent, Value):
            tparent = parent._transform
        else:
            raise TypeError("[%s] Cannot only lookup transforms for type string or ShutTUM.values.Value" % self._sequence)

        tchild = self._transform
        return np.dot(np.linalg.inv(tparent), tchild)

    def __rshift__(self, child):
        if isinstance(child, str):
            if child not in self._sequence._refs:
                raise ValueError("[%s] Cannot find the parent reference %s" % (self._sequence, child))
            tchild = np.array(self._sequence._refs[child]['transform'])
        elif isinstance(child, Value):
            tchild = child._transform
        else:
            raise TypeError("[%s] Cannot only lookup transforms for type string or ShutTUM.values.Value" % self._sequence)

        tparent = self._transform
        return np.dot(np.linalg.inv(tparent), tchild)


class Image(Value):
    r"""
    An image is a :any:`Value` with its reference set to one of ``"cam1"`` ... ``"cam4"``.

    Each image is recorded by a :any:`StereoCamera`, which has a shutter type, so as its images. Note, though,
    that each camera might differ its shutter methods between sequences, to achieve statistical independence. 
    That means, that you cannot rely on e.g. ``"cam1"`` always having ``"global"`` or ``"rolling"`` shutter, nor as the other cams.

    The cameras record data at approximately **20 FPS**, but sometimes their might exist frame drops. 

    You can query a lot of information from an image such as its :any:`shutter <ShutTUM.Image.shutter>`, :any:`exposure <ShutTUM.Image.exposure>` time and :any:`ID <ShutTUM.Image.ID>`.
    Since it is a :any:`Value` all transform shenanigans apply.
    """

    _Vector2 = namedtuple('Vector2', 'x y')
    _DistCoeffs = namedtuple('DistCoeffs', 'k1 k2 r1 r2')

    def __init__(self, stereo, shutter, left):
        self._left = left
        self._stereo = stereo
        self._shutter = shutter

        if shutter not in ['global', 'rolling']:
            raise ValueError('[%s] Shutter type can only be "global" or "rolling"' % stereo._sequence)

        for cam in stereo._sequence._cams:
            # Check if the shutter matches
            if stereo._sequence._cams[cam]['shutter']['type'] != shutter: continue

            # Check if the camera position (left/right) matches the camera name
            if left and cam not in ['cam1', 'cam4']: continue
            if not left and cam not in ['cam2', 'cam3']: continue

            # Now we have found a camera matching the wanted shutter type and position
            super(Image, self).__init__(stereo._sequence, stereo._data[0], cam)
            break  # from any further for loop iteration

        if not p.exists(self.path):
            raise ValueError('[%s] Image "%s" does not exist' % (self._sequence, self.path))

    @property
    def _previous(self):
        return self._stereo._previous

    def __eq__(self, other):
        if not isinstance(other, Image): return False
        return (super(Image, self).__eq__(other)
                and self._left == other._left
                and self._shutter == other._shutter)

    def __str__(self):
        return "Image (%s/#%05d/%.2f)" % (self.reference, self.ID, self.stamp)

    @property
    def resolution(self):
        r""" 
        Returns the resolution of the cameras as a named tuple ``Resolution(width, height)``
        
        .. seealso:: :any:`Sequence.resolution <ShutTUM.Sequence.resolution>`
        """
        return self._sequence.resolution

    @property
    def shutter(self):
        r""" The shutter method with which this image was captured as string, either ``"rolling"`` or ``"global"`` """
        return self._shutter

    @property
    def ID(self):
        r"""The frame ID as int of this image. This number, prepended to 5 digits is also the name of the JPEG file"""
        return self._stereo.ID

    @property
    def opposite(self):
        r"""This holds the reference of the opposite image, that is the image taken from the other camera with the 
        same shutter."""
        return self._stereo._opposite(self)

    @property
    def exposure(self):
        r"""The exposure time as float in milli seconds, that this image was illuminated"""
        return self._stereo.exposure

    @property
    def illuminance(self):
        r"""
        The image's illuminance in lux.
        
        .. seealso:: :any:`StereoImage.illuminance <ShutTUM.StereoImage.illuminance>`
        """
        return self._stereo.illuminance

    @property
    def path(self):
        r"""
        The path to the JPEG file of this image, relative to the the construction parameter of 
        :any:`Sequence(...) <ShutTUM.Sequence.__init__>`.
        """
        return p.join(self._stereo._sequence._path, 'frames', self.reference, '%05d.jpeg' % self.ID)

    @property
    def imu(self):
        r"""The matching :any:`Imu` for this image. Since the capture of an image is synchronized with the IMU,
        no interpolation is needed."""
        return self._stereo.imu

    def groundtruth(self, max_stamp_delta=.5,
                    position_interpolation=ShutTUM.Interpolation.linear,
                    orientation_interpolation=ShutTUM.Interpolation.slerp):
        r"""
        Find the matching :any:`GroundTruth` value for this image. Since the motion capture system and the cameras
        are not synced, we need to interpolate between ground truths by image's time stamp.
        
        :param max_stamp_delta: specify the maximal allowed time deviation [s] for interpolation. When the previous 
                                (or next) closest ground truth value is older (or newer) then this delta, None will be 
                                returned. 
        :param position_interpolation: a predefined or custom interpolation function to interpolate positions 
        :param orientation_interpolation: a predefined or custom interpolation function to interpolate quaternions
        :return: the matching interpolated ground truth or None, if any of its values was NaN

        .. seealso:: :any:`ShutTUM.GroundTruth.interpolate`
        """
        return GroundTruth.interpolate(self._sequence, self.stamp, max_stamp_delta, position_interpolation, orientation_interpolation)

    def load(self):
        r"""
        Loads the JPEG itself into memory::

            image = ...
            pixels = image.load()   # uses cv2.imread()
            print(pixels.shape)     # (1024, 1280)
            print(type(pixels))     # numpy.ndarray

        """
        return cv2.imread(self.path, cv2.IMREAD_GRAYSCALE)

    def distortion(self, model):
        r""" 
        Get the distortion coefficients the camera which took this image
        It looks up the parameters in ``params/params.yaml/<cam>/distortion/<model>``
        
        :param model: One of ``"fov"`` or ``"radtan"``
        :return: either a float :math:`\omega` for FOV model, or a named tuple with 
        the four elements ``DistCooefs(k1, k2, r1, r2)`` for Rad Tan model.
        """
        if model == 'fov':
            return self._sequence._refs[self.reference]['distortion'][model]

        elif model == 'radtan':
            coeffs = self._sequence._refs[self.reference]['distortion'][model]
            return self._DistCoeffs(k1=coeffs[0], k2=coeffs[1], r1=coeffs[2], r2=coeffs[3])

        else:
            raise ValueError("[%s] Unknown distortion model: %s" % (self._sequence, model))

    @property
    def focal(self):
        r""" The camera's focal length as named tuple ``Vector2(x, y)`` in pixels """
        return Image._Vector2(x=self._sequence._refs[self.reference]['intrinsics'][0],
                              y=self._sequence._refs[self.reference]['intrinsics'][1])

    @property
    def principle(self):
        r""" The camera's principle point as named tuple ``Vector2(x, y)`` in pixels """
        return Image._Vector2(x=self._sequence._refs[self.reference]['intrinsics'][2],
                              y=self._sequence._refs[self.reference]['intrinsics'][3])

    @property
    def K(self):
        r"""
        The intrinsic camera matrix as 3x3 
        `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_
        """
        f = self.focal
        p = self.principle
        return np.array(((f.x,   0, p.x),
                         (  0, f.y, p.y),
                         (  0,   0,   1)))


class StereoImage(Value):
    r"""
    A stereo image contains of two individual :any:`Image` s, a left and a right one.
    It is more or less a container for the two.

    Note that for internal reasons a stereo image derives from :any:`Value`. However, you should
    not use the transform functions (``<<`` and ``>>``) with this, since a stereo image contains two 
    reference frames, one for each camera::
        id = 1
        stereo = sequence.cameras('rolling')[id]

        # The following is ambiguous
        notok = stereo << "imu"   # which camera of the two do you mean?

        # Better would be
        ok = stereo.L << "imu"    # (which is the same as above)


    """

    @staticmethod
    def extrapolate(value, shutter, method='closest'):
        r"""
        Find a matching image for a certain :any:`Value` based on a extrapolation method

        :param value: The value for which to find a matching image
        :param shutter: The shutter type of the images to find (``"global"``, ``"rolling"``, **not** ``"both"``) 
        :param method: An optional extrapolation method to determine the rules for a "match":

            * ``"closest"``: the image with the least difference to value.stamp is chosen
            * ``"next"``: the image with the next larger time stamp than value.stamp is chosen
            * ``"prev"``: the image with the next smaller time stamp than value.stamp is chosen
            * ``"exact"``: the image where value.stamp == image.stamp holds is chosen, None otherwise

        :return: The matching stereo image or None if no was found
        
        .. seealso:: :any:`Sequence.cameras <ShutTUM.Sequence.cameras>`
        
        """
        if method == 'closest':
            f = value._sequence.raw.frames
            i = np.abs(f[:, 0] - value.stamp).argmin()
            return StereoImage(value._sequence, f[i, :], shutter)

        if method == 'next':
            f = value._sequence.raw.frames
            frame = f[f[:, 0] > value.stamp, :]
            if frame.size == 0: return None
            return StereoImage(value._sequence, frame[0], shutter)

        if method == 'prev':
            f = value._sequence.raw.frames
            frame = f[f[:, 0] < value.stamp, :]
            if frame.size == 0: return None
            return StereoImage(value._sequence, frame[-1], shutter)

        if method == 'exact':
            f = value._sequence.raw.frames
            frame = f[f[:, 0] == value.stamp, :]
            if frame.size == 0: return None
            return StereoImage(value._sequence, frame[0], shutter)

        raise ValueError('[%s] Unknown extrapolation method: %s (supported are "closest", "next", "prev" and "exact")'
                         % (value._sequence, method))

    def __init__(self, sequence, data, shutter):
        self._data = data
        self._sequence = sequence
        self._shutter = shutter
        L, R = None, None
        frames = sequence.raw.frames
        id = self._data[1]
        while True:
            try: L = Image(self, shutter, left=True)
            except ValueError: L = None
            try: R = Image(self, shutter, left=False)
            except ValueError: R = None

            # If stereosync is required, the stereo image is only valid, if we found both L/R images
            if sequence.stereosync and L is not None and R is not None: break

            # If no stereosync is required, it suffices if we find only one
            if not sequence.stereosync and (L is not None or R is not None): break

            # Otherwise we look for the next data in frames.csv and try again ...
            id += 1
            self._data = frames[frames[:,1] == id]
            if not self._data.size == 0:
                self._data = self._data[0]
                continue

            raise ValueError('[%s] Cannot find no more stereo images due to frame drops for camera "%s"'
                             % (self._sequence, shutter))

        # Timestamp is in second column
        cam = L if L is not None else R
        self._left, self._right = L, R
        super(StereoImage, self).__init__(sequence, self._data[0], cam.reference)

    @property
    def _previous(self):
        return StereoImage.extrapolate(self, self.L.shutter, method='prev')

    def __eq__(self, other):
        if not isinstance(other, StereoImage): return False
        return super(StereoImage, self).__eq__(other) and np.allclose(self._data, other._data)

    def __str__(self):
        l = self._left.reference if self._left is not None else 'None'
        r = self._right.reference if self._right is not None else 'None'
        return "StereoImage ({%s|%s}/%05d" % (l, r, self.ID)

    def _opposite(self, img):
        if img is self._left: return self._right
        if img is self._right: return self._left
        raise ValueError("[%s] Image %s unknown, cannot find opposite" % self._sequence)

    @property
    def resolution(self):
        r""" 
        Returns the resolution of the cameras as a named tuple ``Resolution(width, height)``
        
        .. seealso:: :any:`Sequence.resolution <ShutTUM.Sequence.resolution>`
        
        """
        return self._sequence.resolution

    @property
    def ID(self):
        r"""The frame ID as int of this image. This number, prepended to 5 digits is also the name of the JPEG file"""
        return int(self._data[1])

    @property
    def shutter(self):
        r""" The shutter method with which this stereo image was captured as string, either ``"rolling"`` or ``"global"`` """
        return self._shutter

    @property
    def exposure(self):
        r"""
        The exposure time as float in milli seconds, that this image was illuminated. This is constant for 
        :any:`L <ShutTUM.StereoImage.L>` and :any:`R <ShutTUM.StereoImage.R>`
        """
        return self._data[2]

    @property
    def illuminance(self):
        r"""
        The estimated illuminance measured by the `TSL2561 Lux-sensor <https://cdn-shop.adafruit.com/datasheets/TSL2561.pdf>`_. 
        
        This float is measured in lx but is only an approximation. This value is constant for both 
        :any:`L <ShutTUM.StereoImage.L>` and :any:`R <ShutTUM.StereoImage.R>`. Based on this value, the estimated 
        :any:`exposure` time can be computed with the following hand fitted formula:

        .. image:: images/autoexposure.svg

        .. math:: t_{exposure} \left ( E_v \right ) = 198.1527 \cdot E_v^{-0.8003}
        
        This will return ``float('nan')`` if no measurement was taken, such as during an I2C error.
        """
        return self._data[3]

    @property
    def L(self):
        r""" The reference to the left :any:`Image`"""
        return self._left

    @property
    def R(self):
        r""" The reference to the right :any:`Image`"""
        return self._right

    @property
    def imu(self):
        r"""The matching :any:`Imu` for this image. Since the capture of an image is synchronized with the IMU,
                no interpolation is needed."""
        i = self._sequence.raw.imu
        match = i[i[:, 0] == self.stamp]
        if match.size > 0: return Imu(self._sequence, match[0])

        raise ValueError("[%s] It seems that %s has no matching IMU value" % (self._sequence, self))


class Imu(Value):
    r""" An imu value represents the measurement of IMU sensor at a specific time.

    Since it is a :any:`Value` you can use it to calculate transforms with it. Also 
    the IMU is synchronized in a way, that it measures exactly seven times 
    per image. Any :any:`Imu` value consist of three acceleration measurements in X, Y and Z 
    and three angular velocity measurements around the X, Y, and Z axis.

    .. seealso:: :any:`Interpolation` for the data frequency & :any:`Value` for the reference frame

    """

    @staticmethod
    def extrapolate(value, method='closest'):
        r"""
        Find a matching :any:`Imu` value for a certain :any:`Value` based on a extrapolation method

        :param value: The value for which to find a matching :any:`Imu` value
        :param method: An optional extrapolation method to determine the rules for a "match":

            * ``"closest"``: the image with the least difference to value.stamp is chosen
            * ``"next"``: the image with the next larger time stamp than value.stamp is chosen
            * ``"prev"``: the image with the next smaller time stamp than value.stamp is chosen
            * ``"exact"``: the image where value.stamp == image.stamp holds is chosen, None otherwise

        :return: The matching IMU value or None if no was found
        """
        if method == 'closest':
            f = value._sequence.raw.imu
            i = np.abs(f[:, 0] - value.stamp).argmin()
            return Imu(value._sequence, f[i, :])

        if method == 'next':
            f = value._sequence.raw.imu
            imu = f[f[:, 0] > value.stamp, :]
            if imu.size == 0: return None
            return Imu(value._sequence, imu[0])

        if method == 'prev':
            f = value._sequence.raw.imu
            imu = f[f[:, 0] < value.stamp, :]
            if imu.size == 0: return None
            return Imu(value._sequence, imu[-1])

        if method == 'exact':
            f = value._sequence.raw.imu
            imu = f[f[:, 0] == value.stamp, :]
            if imu.size == 0: return None
            return Imu(value._sequence, imu[0])

        raise ValueError('[%s] Unknown extrapolation method: %s (supported are "closest", "next", "prev" and "exact")'
                         % (value._sequence, method))

    @staticmethod
    def interpolate(sequence, stamp, acceleration_interpolation=ShutTUM.Interpolation.linear,
                    angvelo_interpolation=ShutTUM.Interpolation.linear):
        r"""
        This function enables you to find the interpolated :any:`Imu` values of a record given a certain timestamp.

        :param sequence: the sequence which holds all imu values to interpolate over 
        :param float stamp: the time at which to interpolate (in seconds, with decimal places) 
        :param acceleration_interpolation: A predefined or custom interpolation function
        :param angvelo_interpolation: A predefined or custom interpolation function
        :return: An :any:`Imu` value

        .. seealso:: :any:`Interpolation`
        """

        imu = sequence.raw.imu
        idx = np.searchsorted(imu[:, 0], stamp)
        acc = np.ones((1, 3)) * np.nan
        gyr = np.ones((1, 4)) * np.nan
        if idx != 0 and idx != imu.shape[0]:
            ta = acceleration_interpolation(imu[idx - 1, 0], imu[idx, 0], stamp)
            acc = (1 - ta) * imu[idx - 1, 1:4] + ta * imu[idx, 1:4]
            tg = angvelo_interpolation(imu[idx - 1, 0], imu[idx, 0], stamp)
            gyr = (1 - tg) * imu[idx - 1, 1:4] + tg * imu[idx, 1:4]

        return Imu(sequence, np.concatenate(([stamp], acc, gyr)))

    def __init__(self, sequence, data):
        if len(data) < 7: raise ValueError(
            "[%s] Data must have at least 7 entries [time1, acc3, gyro3] but has %d" % (self._sequence, len(data)))
        super(Imu, self).__init__(sequence, stamp=data[0], reference='imu')
        self._acc = np.array(data[1:4])
        self._gyro = np.array(data[4:7])

    @property
    def _previous(self):
        return Imu.extrapolate(self, method='prev')

    def __eq__(self, other):
        if not isinstance(other, Imu): return False
        return (super(Imu, self).__eq__(other)
                and np.allclose(self._acc, other._acc)
                and np.allclose(self._gyro, other._gyro))

    @property
    def acceleration(self):
        r"""
        The acceleration 3D vector [x,y,z] of this measurement as 
        `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_ 
        in :math:`\frac{m}{s^2}`
        """
        return self._acc

    @property
    def angular_velocity(self):
        r"""
        The angular velocity 3D vector around [x,y,z] of this measurement as 
        `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_
        in :math:`\frac{rad}{s}`
        """
        return self._gyro

    def stereo(self, shutter, extrapolation='closest'):
        r"""
        The matching stereo image for this :any:`Imu` measurement
        :param shutter: The shutter type of the images to find (``"global"``, ``"rolling"``, **not** ``"both"``) 
        :param extrapolation: An optional extrapolation method to determine the rules for a "match":

            * ``"closest"``: the image with the least difference to value.stamp is chosen
            * ``"next"``: the image with the next larger time stamp than value.stamp is chosen
            * ``"prev"``: the image with the next smaller time stamp than value.stamp is chosen
            * ``"exact"``: the image where ``value.stamp == image.stamp`` holds is chosen, None otherwise
        :return: The matching stereo image or None if no was found

        .. seealso:: :any:`ShutTUM.StereoImage.extrapolate`
        
        """
        return StereoImage.extrapolate(self, shutter, method=extrapolation)

    def groundtruth(self, max_stamp_delta=.5,
                    position_interpolation=ShutTUM.Interpolation.linear,
                    orientation_interpolation=ShutTUM.Interpolation.slerp):
        r"""
        Find the matching :any:`GroundTruth` value for this :any:`Imu` value. Since the motion capture system and the 
        IMU sensor are not synced, we need to interpolate between ground truths by time stamp of this :any:`Imu` value.

        :param max_stamp_delta: specify the maximal allowed time deviation [s] for interpolation. When the previous 
                                (or next) closest ground truth value is older (or newer) then this delta, None will be 
                                returned. 
        :param position_interpolation: a predefined or custom interpolation function to interpolate positions 
        :param orientation_interpolation: a predefined or custom interpolation function to interpolate quaternions
        :return: the matching interpolated ground truth

        .. seealso:: :any:`ShutTUM.GroundTruth.interpolate`
        
        """
        return GroundTruth.interpolate(self._sequence, self.stamp, max_stamp_delta,
                                       position_interpolation, orientation_interpolation)


class GroundTruth(Value):
    r"""
    A ground truth is a :any:`Value` with the reference ``"world"``.
    The ground truth is taken with a higher frequency than the images (around 120 Hz), but slower than the imu. 
    Since the mocap system is stationary in one room only, it will not cover the whole duration of this sequence.
    """

    @staticmethod
    def interpolate(sequence, stamp, max_stamp_delta=.5,
                    position_interpolation=ShutTUM.Interpolation.linear,
                    orientation_interpolation=ShutTUM.Interpolation.slerp):
        r"""
        This function enables you to find the interpolated ground truth of a record given a certain timestamp.

        :param sequence: the sequence which holds all ground truth values to interpolate over 
        :param float stamp: the time at which to interpolate (in seconds, with decimal places) 
        :param max_stamp_delta: specify the maximal allowed time deviation [s] for interpolation. When the previous 
                                (or next) closest ground truth value is older (or newer) then this delta, None will be 
                                returned. 
        :param position_interpolation: A predefined or custom interpolation function
        :param orientation_interpolation: A predefined or custom interpolation function
        :return: A :any:`GroundTruth` -Value

        .. seealso:: :any:`Interpolation`
        """
        poses = sequence.raw.groundtruth
        idx = np.searchsorted(poses[:, 0], stamp)
        p = np.ones((1, 3)) * np.nan
        q = np.ones((1, 4)) * np.nan
        if idx != 0 and idx != poses.shape[0]:
            dt = stamp - poses[idx -1, 0]
            if dt > max_stamp_delta: return None

            t = dt / (poses[idx, 0] - poses[idx - 1, 0])

            # Compute previous values for interpolation method, if existant
            a0, b0 = np.zeros((3, 1)), np.zeros((3, 1))
            qa0, qb0 = np.array((1, 0, 0, 0)), np.array((1, 0, 0, 0))
            if idx - 1 >= 0:             a0, qa0 = poses[idx - 2, 1:4], poses[idx - 2, 4:8]
            if idx + 1 < poses.shape[0]: b0, qb0 = poses[idx + 1, 1:4], poses[idx + 1, 4:8]

            p = position_interpolation(poses[idx - 1, 1:4], poses[idx, 1:4], t, a0, b0)
            q = orientation_interpolation(poses[idx - 1, 4:8], poses[idx, 4:8], t, qa0, qb0)
            if np.any(np.isnan(p)) or np.any(np.isnan(q)): return None

        return GroundTruth(sequence, np.concatenate(([stamp], p, q)))

    @staticmethod
    def extrapolate(value, method='closest'):
        r"""
        Find a matching ground truth for a certain :any:`Value` based on a extrapolation method

        :param value: The value for which to find a matching ground truth value
        :param method: An optional extrapolation method to determine the rules for a "match":

            * ``"closest"``: the image with the least difference to value.stamp is chosen
            * ``"next"``: the image with the next larger time stamp than value.stamp is chosen
            * ``"prev"``: the image with the next smaller time stamp than value.stamp is chosen
            * ``"exact"``: the image where value.stamp == image.stamp holds is chosen, None otherwise

        :return: The matching ground truth or None if no was found
        """
        if method == 'closest':
            f = value._sequence.raw.groundtruth
            i = np.abs(f[:, 0] - value.stamp).argmin()
            return GroundTruth(value._sequence, f[i, :])

        if method == 'next':
            f = value._sequence.raw.groundtruth
            gt = f[f[:, 0] > value.stamp, :]
            if gt.size == 0: return None
            return GroundTruth(value._sequence, gt[0])

        if method == 'prev':
            f = value._sequence.raw.groundtruth
            gt = f[f[:, 0] < value.stamp, :]
            if gt.size == 0: return None
            return GroundTruth(value._sequence, gt[-1])

        if method == 'exact':
            f = value._sequence.raw.groundtruth
            gt = f[f[:, 0] == value.stamp, :]
            if gt.size == 0: return None
            return GroundTruth(value._sequence, gt[0])

        raise ValueError('[%s] Unknown extrapolation method: %s (supported are "closest", "next", "prev" and "exact")'
                         % (value._sequence, method))

    def __init__(self, sequence, data):
        if len(data) < 8:
            raise ValueError("[%s] Data must have at least 8 entries [time1, position3, orientation4] but has %d" %
                             (sequence, len(data)))
        self._data = data
        self._sequence = sequence
        super(GroundTruth, self).__init__(sequence, stamp=data[0], reference='world')

    @property
    def _previous(self):
        return GroundTruth.extrapolate(self, method='prev')

    def __eq__(self, other):
        if not isinstance(other, GroundTruth): return False
        return super(GroundTruth, self).__eq__(other) and np.allclose(self._data, other._data)

    @property
    def _transform(self):
        return tf.affines.compose(
            T=self._data[1:4],
            R=tf.quaternions.quat2mat(self._data[4:8]),
            Z=np.ones(3),
            S=np.zeros(3)
        ).dot(self.marker >> 'cam1')

    def stereo(self, shutter, extrapolation='closest'):
        r"""
        Find a matching stereo image pair for this ground truth value.

        :param shutter: The shutter type of the images to find (``"global"``, ``"rolling"``, **not** ``"both"``) 
        :param extrapolation: An optional extrapolation method to determine the rules for a "match" 
                            (one of ``"closest"``, ``"next"``, ``"prev"``, ``"exact"``)
        :return: The matching stereo image or None if no was found

        .. seealso:: :any:`ShutTUM.StereoImage.extrapolate`
        
        """
        return StereoImage.extrapolate(self, shutter, method=extrapolation)

    @property
    def marker(self):
        r""" Returns the :any:`Marker`, which got tracked by this ground truth. """
        return Marker(self._sequence, self.stamp)


class Marker(Value):
    r"""
    The marker is an IR ball placed on top of the setup to be tracked by the mocap.
    
    To enable transformation calculations, this class inherits from :any:`Value`,
    though its transformation stays constant over time.
    
    .. seealso:: :any:`Value`
    
    """
    def __init__(self, sequence, stamp):
        super(Marker, self).__init__(sequence, stamp, reference='marker')



import yaml
import cv2
import os.path as p
import numpy   as np

from collections import namedtuple
import ShutTUM.devices
import ShutTUM.values


class Sequence(object):
    r"""The base class representing one dataset sequence.
    
    In order for performent operations the sequence loads and checks lots of the data in its :any:`__init__` thus avoiding
    expensive checks in for loops or list comprehensions. In addition it holds the reference to list-like objects, such 
    as :any:`cameras` or :any:`imu <ShutTUM.Sequence.imu>`. You can iterate over each of these (depending
    on application) to get the corresponding :any:`Value` s in order.
    
    A typical application might look like::
    
        # Load a dataset's sequence
        sequence = Sequence('path/to/folder/01')
        
        # Iterate over all imu values and find corresponding images
        for observation in sequence.imu:
            
            print(observation.acceleration)
            print(observation.angular_velocity)
            
            stereo = observation.stereo(shutter='global')
            if stereo is None: continue
            
            print(stereo.L.ID)
    
    """

    _Data = namedtuple('Data', 'global_ rolling imu groundtruth stamp')

    @staticmethod
    def _check_folder_exists(folder):
        r""" Checks if a folder exists and raises an exception otherwise
        :param str folder: the path to the folder to check for existance
        :raises IOError
        """
        if not p.exists(folder):
            raise IOError("Could not find folder %s, record folder seems not to be valid!" % folder)

    @staticmethod
    def _check_file_exists(file):
        r""" Checks if a file exists and raises an exception otherwise
        :param str file: the path to the file to check for existance
        :raises IOError
        """
        if not p.exists(file):
            raise IOError("Could not find %s, record folder seems not to be valid!" % file)

    @staticmethod
    def _check_contains_key(file, dictionary, key):
        r"""
        :param str file: The reference file, which to give in the error message
        :param dict dictionary: the dict to check the key's existance
        :param str key: the key to check for 
        :raises ValueError
        
        Checks if a dict originating from a certain file contains a certain key and raises an exception otherwise
        """
        if key not in dictionary:
            raise ValueError("Could not find %s in %s, record folder seems not to be valid!" % (key, file))

    def __init__(self, path, stereosync=True, calibration=False):
        r""" 
        :param str path: the path to one sequence of the dataset, such as ``~/ShutTUM/0001``
        :param bool stereosync: possiblity to set the :any:`stereosync` option in constructor
        :param bool calibration: is this a calibration sequence? If yes, ``cam1`` and ``cam2`` are used as ``global`` 
        while ``cam3`` and ``cam4`` are set under the ``rolling`` namespace 
        :raises: ValueError: if anything goes wrong 
        Load the sequence into memory (except images) and do basic data consistency checks.
        
        1. It is checked that in path there exists a ``data``, ``frames`` and ``params`` folder
        2. It is checked that there exists the files ``data/frames.csv``, ``data/imu.csv`` and ``data/ground_truth.csv``
        3. The files from 2. are loaded into memory (see :any:`raw`)
        4. The ``params/time.yaml`` is loaded
        5. The ``params/params.yaml`` is loaded
           
        .. seealso:: :any:`stereosync`
        
        """
        path = p.expandvars(p.expanduser(path))
        self._path = path
        self._sync = stereosync

        # Consistency Check
        Sequence._check_folder_exists(p.join(path, 'data'))
        Sequence._check_folder_exists(p.join(path, 'frames'))
        Sequence._check_folder_exists(p.join(path, 'params'))

        f = p.join(path, 'data', 'frames.csv')
        i = p.join(path, 'data', 'imu.csv')
        g = p.join(path, 'data', 'ground_truth.csv')

        Sequence._check_file_exists(f)
        Sequence._check_file_exists(i)
        Sequence._check_file_exists(g)

        self._frames       = np.genfromtxt(f, delimiter='\t', skip_header=1)#, dtype=(float, int, float, str))
        self._imu          = np.genfromtxt(i, delimiter='\t', skip_header=1)
        self._ground_truth = np.genfromtxt(g, delimiter='\t', skip_header=1)
        self._times = list(sorted(set(self._frames[:,0]) | set(self._imu[:,0]) | set(self._ground_truth[:,0])))

        Raw = namedtuple('Raw', ['frames', 'imu', 'groundtruth'])
        self._raw = Raw(self._frames, self._imu, self._ground_truth)
        self._time = {}
        timefile = p.join(path, 'params', 'time.yaml')
        Sequence._check_file_exists(timefile)
        with open(timefile) as stream:
            self._time = yaml.load(stream)['time']
            Sequence._check_contains_key(timefile, self._time, 'start')
            Sequence._check_contains_key(timefile, self._time, 'end')
            Sequence._check_contains_key(timefile, self._time, 'duration')

        self._cams = {}
        paramfile = p.join(path, 'params', 'params.yaml')
        Sequence._check_file_exists(paramfile)
        with open(paramfile) as stream:
            self._refs = yaml.load(stream)
            if calibration:
                self._refs['cam1']['shutter']['type'] = 'global'
                self._refs['cam2']['shutter']['type'] = 'global'
                self._refs['cam3']['shutter']['type'] = 'rolling'
                self._refs['cam4']['shutter']['type'] = 'rolling'
                # TODO add doc string for this

        self._gammas = {}
        self._cams = {}
        for ref in self._refs:
            if ref == 'world': continue  # world param must only be present, not more

            # Every other reference must have at least a transform parameter
            Sequence._check_contains_key(paramfile, self._refs[ref], 'transform')

            if 'shutter' not in self._refs[ref]: continue

            # If the reference contains a 'shutter' param, it is a camera
            cam = ref
            self._cams[cam] = self._refs[ref]

            folder   = p.join(path,   'params', cam)
            gamma    = p.join(folder, 'gamma.txt')
            vignette = p.join(folder, 'vignette.png')
            Sequence._check_folder_exists(folder)
            Sequence._check_file_exists(gamma)
            Sequence._check_file_exists(vignette)
            self._gammas[cam] = np.genfromtxt(gamma, delimiter=' ')

        self._cameras = ShutTUM.devices.DuoStereoCamera(self)

    def __str__(self):
        return "%s (%s)" % (type(self).__name__, p.basename(p.normpath(self._path)))

    @property
    def path(self):
        r"""The path to this sequence. Environment variables get expanded automatically"""
        return self._path

    @property
    def raw(self):
        r"""
        The raw values in matrix form. This property is a Named Tuple with the following fields:
        
        * ``frames`` (`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) corresponding to ``data/frames.csv``
        * ``imu`` (`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) corresponding to ``data/imu.csv``
        * ``groundtruth`` (`ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_) corresponding to ``data/ground_truth.csv``
        
        """
        return self._raw

    @property
    def stereosync(self):
        r"""
        Enables/Disables frame synchronization between cameras. (default ``True``)
        
        When working with stereo images you obviously need two images. Sometimes, however, one camera has recorded a 
        frame while its :any:`opposite` has dropped a frame. With the ``stereosync`` option you can specify how to 
        handle these frame drops. 
             
        .. image:: images/stereosync.svg
        
        With ``stereosync`` enabled, you iterate only over frames which have been captured by *both* cameras. With
        stereosync disabled, you iterate over all stereo images which have *at least one* camera captured an image. If both 
        cams occurred to drop the same frame, this will be skipped in the iterations. The dropped :any:`Image` is set
        to None::
        
            sequence.stereosync = False
            
            left_drops, right_drops = [], []
            for stereo in sequence.cameras('global'):
                if stereo.L is None: left_drops.append(stereo.ID)
                if stereo.R is None: right_drops.append(stereo.ID)
                
            print('Left Camera dropped frames:  %s' % left_drops)
            print('Right Camera dropped frames: %s' % right_drops)
            
        """
        return self._sync

    @stereosync.setter
    def stereosync(self, value):
        self._sync = value

    def cameras(self, shutter='both'):
        r""" 
        :param str shutter: {both/global/rolling} the type of shutter you are interested in.
        :return: The reference of the cameras, which you can iterate either as :any:`StereoCamera` (global/rolling)or :any:`DuoStereoCamera` (both)
        :raise: ValueError: for anything other then both/global/rolling
        
        Get a reference to one or both of the two stereo cameras, to iterate over their images
        
        Note that often you want to directly iterate over those references, which you can do like::
        
            # Iterate over both rolling and global images
            for g, r in sequence.cameras(shutter='both'):
                print(g.L.ID)
                print(r.R.stamp)
        
        
        .. seealso:: :any:`ShutTUM.Sequence.stereosync`
       
        """
        if shutter == 'both':    return self._cameras
        if shutter == 'global':  return self._cameras._global
        if shutter == 'rolling': return self._cameras._rolling

        raise ValueError('Unknown shutter type: use either "global", "rolling", or "both" and not %s' % shutter)

    @property
    def imu(self):
        r""" 
        :return: Generator property, yielding one :any:`Imu` after the other
        
        This property is a python generator, which you can use to iterate over all available
        :any:`Imu` s in this sequence. Note that generators do not support indexing. 
        
        Note that often you want to directly iterate over the Imu like so::
            
            # Iterate over all imu values
            for observation in sequence.imu:
                print(observation.acceleration)
                
            # When you want e.g. the first imu value use:
            first_imu = next(sequence.imu)
            
            # When you absolutely need indexing convert the generator to a in memory list first
            imus = list(sequence.imu)
            print(imus[17].stamp)
            
            # Use the usual python "functional" shenanigans
            imus_up_to_10s = filter(lambda imu: imu.stamp <= 10, sequence.imu)
            imus_up_to_10s = [ imu for imu in sequence.imu if imu.stamp <= 10 ]
            
            acc_in_g = map(lambda imu: imu.acceleration * 9.805, sequence.imu)
            acc_in_g = [ imu.acceleration * 9.805 for imu in sequence.imu ]
            
            # Note that generators don't implement __len__()
            # to get the amount of imu values in this sequence either use
            N = len(list(sequence.imu))               # or:
            N = sum([ 1 for value in sequence.imu ])  # which is the same as:
            N = sequence.raw.imu.shape[0]
        
        """
        for row in self.raw.imu:
            yield ShutTUM.values.Imu(self, row)

    @property
    def mocap(self):
        r""" 
        :return: Generator property, yielding one :any:`GroundTruth` value after the other

        This property is a python generator, which you can use to iterate over all available
        :any:`GroundTruth` s in this sequence. Note that generators do not support indexing. 

        Note that often you want to directly iterate over the Mocap like so::

           # Iterate over all ground truth values
           for gt in sequence.mocap:
               c1_T_w = gt >> 'cam1'   # pose of cam1 in world frame

           # When you want e.g. the first ground truth value use:
           first_gt = next(sequence.mocap)

           # When you absolutely need indexing convert the generator to a in memory list first
           gts = list(sequence.mocap)
           print(gts[17].stamp)

           # Use the usual python "functional" shenanigans
           gts_from_10s = filter(lambda gt: gt.stamp >= 10, sequence.mocap)
           gts_from_10s = [ gt for gt in sequence.mocap if gt.stamp >= 10 ]

           # Note that generators don't implement __len__()
           # to get the amount of imu values in this sequence either use
           N = len(list(sequence.mocap))               # or:
           N = sum([ 1 for value in sequence.mocap ])  # which is the same as:
           N = sequence.raw.groundtruth.shape[0]

        """
        for row in self.raw.groundtruth:
            yield ShutTUM.values.GroundTruth(self, row)

    @property
    def times(self):
        r"""
        A list of all time stamps in this sequence.
        
        .. math:: \mathbf{t} = \mathbf{t}_{frames} \cup \mathbf{t}_{imu} \cup \mathbf{t}_{groundtruth}
        
        Note that this list is sorted, so you can easily iterate over it like so::
        
            for time in sequence.times:
                print(time)
        
        
        """
        return self._times

    @property
    def start(self):
        r"""
        The start time of the record as Unix Timestamp in seconds with decimal places
        """
        return self._time['start']

    @property
    def end(self):
        r"""
        The end time of the record as Unix Timestamp in seconds with decimal places
        """
        return self._time['end']

    @property
    def duration(self):
        r""" 
        The duration of the record in seconds as float, so basically:
        ``sequence.end - sequence.start``
        
        """
        return self._time['duration']

    @property
    def resolution(self):
        r""" Returns the resolution of the cameras as a named tuple ``Resolution (width, height)`` """
        Resolution = namedtuple('Resolution', 'width height')
        return Resolution(width=1280, height=1024)

    @property
    def exposure_limits(self):
        r"""
        :return: a Namped Tuple ``Limits(min, max)`` indicating the minimum
        and maximum exposure time in milliseconds which occurred during this sequence
        
        The minimal & maximal exposure used for all cameras. Note that these values are the *limits*
        not the extrema of the record, so most of the time, these will not be reached, but if, clamped accordingly.::
        
            limits = sequence.exposure_limits
            print("Limits are %s .. %s ms" % (limits.min, limits.max))
            
        
        """
        Limits = namedtuple('Limits', ['min', 'max'])
        for cam in self._cams:
            # take the first camera, since all limits are the same
            exp = self._cams[cam]['exposure']
            return Limits(min=exp['min'], max=exp['max'])

    def gamma(self, cam, input):
        r""" 
        :param str cam: the name of the camera (e.g. ``"cam1"``)
        :param float input: the position to lookup, i.e. X-axis on luminance plot. Between 0 .. 255, will be rounded to int
        :raises: ValueError: for unknown camera names or inputs below 0 or above 255
        
        Lookup a gamma value from ``params/<cam>/gamma.txt``
        """
        if cam not in self._cams:
            raise ValueError("Unknown camera name: %s" % cam)
        if input < 0 or input > 255:
            raise ValueError("Gamma function only defined for inputs from 0 .. 255 and not for %s" % input)
        return self._gammas[cam][int(round(input))]

    def vignette(self, cam):
        r"""
        :param str cam: the name of the camera to lookup its vignette (e.g. ``"cam1"``)  
        :return: the vignette image, read by `cv2.imread() <http://docs.opencv.org/3.0-beta/doc/py_tutorials/py_gui/py_image_display/py_image_display.html>`_ with dimensions [1280x1024] as grayscale
        :rtype: `ndarray <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.ndarray.html>`_
        """
        if cam not in self._cams:
            raise ValueError("Unknown camera name: %s" % cam)
        file = p.join(self._path, 'params', cam, 'vignette.png')
        return cv2.imread(file, cv2.IMREAD_GRAYSCALE)

    @property
    def rolling_shutter_speed(self):
        r"""
        How fast did the two rolling shutter cameras shuttered. Returns the time between the exposure of two consecutive 
        rows in milli seconds (approximate)
        
        """
        for cam in self._cams:
            shutter = self._cams[cam]['shutter']
            if shutter['type'] == 'rolling':
                return shutter['speed']
        raise ValueError('No cams in %s had rolling shutter enabled!' % self._path)

    @property
    def shutter_types(self):
        r"""
        Which shutter method uses each camera.
        
        Returns one of the two following dictionaries::
        
            # For sequences with odd ID
            {
                'cam1': 'global'
                'cam2': 'global'
                'cam3': 'rolling'
                'cam4': 'rolling'
            }
            
            # For sequences with even ID
            {
                'cam1': 'rolling'
                'cam2': 'rolling'
                'cam3': 'global'
                'cam4': 'global'
            }
            
        """
        return { cam: self._cams[cam]['shutter']['type'] for cam in self._cams }

    def lookup_cam_name(self, shutter, side):
        r""" Find the corresponding name of the camera for a given shutter and side
        
        :param shutter: either one of {``"global"``, ``"rolling"``} 
        :param side: either one of {``'L'``, ``'R'`` }
        :return: one of {``"cam1"`` .. ``"cam4"``} or raises ValueError on invalid parameters
        """
        for name in self._cams:
            if self._cams[name]['shutter']['type'] != shutter: continue
            if side == 'L' and name in ['cam1', 'cam4']: return name
            if side == 'R' and name in ['cam2', 'cam3']: return name

        raise ValueError('%s Cannot lookup cam name for shutter "%s" on side "%s"' % (self, shutter, side))

    def _find_data_for(self, s):
        value = ShutTUM.values.Value(self, s, 'world')  # world as dummy for the time stamp
        return Sequence._Data(
            stamp=s,
            global_=ShutTUM.values.StereoImage.extrapolate(value, 'global', method='exact'),
            rolling=ShutTUM.values.StereoImage.extrapolate(value, 'rolling', method='exact'),
            imu=ShutTUM.values.Imu.extrapolate(value, method='exact'),
            groundtruth=ShutTUM.values.GroundTruth.extrapolate(value, method='exact')
        )

    def _find_data_between(self, start, stop):
        for time in self._times:
            if start is not None and time < start: continue
            if stop  is not None and time > stop:  continue
            yield self._find_data_for(time)

    def __getitem__(self, stamp):
        r"""
        :param float/slice stamp: either the time at which to look up the values or a slice object (e.g. start:stop) 
                                  defining a range of time stamps, between which to lookup the values. If the start 
                                  value of the slice is before :any:`start`, or the end value of the slice is after 
                                  :any:`end`, then the generator yield up to :any:`start` or :any:`end`, respectively.
                                  Note that times for the slice are both *inclusive* unlike normal python index slices.
        :return: either a named tuple with the fields 
                  * ``stamp`` (float)
                  * ``global_`` (:any:`StereoImage`) 
                  * ``rolling`` (:any:`StereoImage`) 
                  * ``imu`` (:any:`Imu`) and 
                  * ``groundtruth`` (:any:`GroundTruth`). 
                 
                 or a generator yielding multiple (or none) of these. If any of the values of the above tuple does not 
                 exist for the stamp, it becomes ``None``. Note the spelling of ``global_``, since ``global`` 
                 is a reserved keyword in python.
        
        Looks up all corresponding :any:`Value` s it can find for a given time stamp::
        
            sequence = Sequence(...)
            
            # The single lookup with one float as index
            for time in sequence.times:
                data = sequence[time]
                if data.global_     is not None: print(data.global_.ID)
                if data.rolling     is not None: print(data.rolling.ID)
                if data.imu         is not None: print(data.imu.acceleration)
                if data.groundtruth is not None: print(data.groundtruth.marker)
        
            # ... or the sliced version specifying all data between 5s .. 45s
            for data in sequence[5:45]:
               print(data)
                
            # ... or all up to 10s
            for data in sequence[:10]:
                print(data)
                
            # ... or all from 30s till end
            for data in sequence[30:]:
                print(data)
                
            # Custom steps, however, are not supported:
            try:
                x = sequence[::-1]
            except ValueError:
                print("Doesn't make sense...")
        
                
            
        Note, though, that this is not the most performant thing to iterate over the sequence,
        since the lookup has to be done on every iteration. Consider iterating over :any:`cameras`, :any:`imu` or 
        :any:`mocap` instead.
        """
        if isinstance(stamp, slice):
            if stamp.step is not None:
                raise ValueError('[%s] Slicing a sequence with a step value like %s:%s:%s is not supported'
                                 % (self, stamp.start, stamp.stop, stamp.step))
            return self._find_data_between(stamp.start, stamp.stop)
        else:
            return self._find_data_for(stamp)




import ShutTUM.values


class StereoCamera(object):
    r"""
    A stereo camera consists of a left and a right camera facing both in the same direction. Both cameras have the same
    shutter method, so either rolling or global. If you iterate over a stereo camera you go through all the 
    :any:`StereoImage` in order::
        
        # You can use a "classic" for loop ... 
        for stereo in sequence.cameras('global'):
            print(stereo.stamp)
        
        # Or filter/map functions ...
        stereo = filter(lambda item: item.ID == 2, sequence.cameras('rolling'))
        print(list(stereo))
        
        # Or even list & dict comprehensions
        stereo = [ image.stamp for image in sequence.cameras('rolling') if image.ID == 2]
        print(stereo)
    
    """
    def __init__(self, sequence, shutter):
        r"""
        Creates a new StereoCamera as iterable container for images. Usually you will not invoke this constructor
        directly but rather get a reference via :any:`cameras <ShutTUM.Sequence.cameras>`
        
        :param Sequence sequence: The reference to the sequence 
        :param str shutter: the name of the shutter this camera uses (usually "rolling" or "global")
        """
        self._sequence = sequence
        self._shutter = shutter
        self._data = self._sequence.raw.frames

    def __iter__(self):
        self._id = int(self._data[0,1])   # take the first ID (column 1) in the first row (0th)
        self._max = int(self._data[-1,1]) # take the last ID (column 1) in the last row
        return self

    def next(self): # python2 support
        return self.__next__()

    def __next__(self):
        while True:
            try:
                img = self[self._id]
                self._id = img.ID + 1
                return img

            except ValueError:
                self._id = int(self._data[0, 1])  # need to reset so list(...) works again
                raise StopIteration

    def __len__(self):
        return sum([1 for frame in self])

    def __getitem__(self, id):
        row = self._data[self._data[:,1] == id, :]
        if row.size == 0: raise ValueError('[%s] No Frame for stereo camera "%s" found with ID %d' % (self._sequence, self._shutter, id))
        return ShutTUM.values.StereoImage(self._sequence, row[0], self._shutter)


class DuoStereoCamera:
    r"""
    A duo stereo camera consists of a two :any:`StereoCamera` s each with a left and a right one. Both stereo cameras
    use usually (but not necessarily) different shutter methods. If you iterate over a duo stereo camera you go get a 
    tuple with both :any:`StereoImage` s in order::

        # You can use a "classic" for loop ... 
        for stereo_1, stereo_2 in sequence.cameras():     # both is the default
            print(stereo_1.shutter)
            print(stereo_2.shutter)

    """
    def __init__(self, sequence):
        self._rolling = StereoCamera(sequence, 'rolling')
        self._global  = StereoCamera(sequence, 'global')

    def __iter__(self):
        self._rolling.__iter__()
        self._global.__iter__()
        return self

    def next(self): # python2 support
        return self.__next__()

    def __next__(self):
        g = self._global.__next__()
        r = self._rolling.__next__()
        return g, r

    def __len__(self):
        return min(len(self._global), len(self._rolling))

    def __getitem__(self, item):
        return self._global[item], self._rolling[item]



from setuptools import setup, find_packages

setup(
    name='ShutTUM',
    version='1.0',
    author='Thore Goll',
    author_email='thoregoll@googlemail.de',
    description='A utility API to easily interact with the ShutTUM dataset',
    url='https://github.com/gollth/ShutTUM',
    license='MIT',
    packages=['ShutTUM'],
    keywords='dataset rolling shutter visual odometry slam',
    long_description=open('README.md').read(),
    install_requires=[
        'opencv-python>=3.2.0',
        'numpy>=1.12',
        'transforms3d>=0.3.1',
        'PyYAML>=3.12'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Localization',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
      ]
)